#!/usr/bin/python
# -*- coding: utf-8 -*-

# Platforms tested:
# Mac OSX 10.4, Python 2.3.5
# Mac OSX 10.6, Python 2.6, Python 3.4
# Windows XP,   Python 2.7

# Version-specific imports
try:
    from Tkinter import *
    from tkFileDialog import askopenfilename,asksaveasfilename
    from tkFont import Font
    from tkMessageBox import askokcancel,askyesno,showerror
except ImportError:
    from tkinter import *
    from tkinter.filedialog import askopenfilename,asksaveasfilename
    from tkinter.font import Font
    from tkinter.messagebox import askokcancel,askyesno,showerror
    
from getopt import getopt,GetoptError
import sys
from os.path import basename
from time import sleep

from kanren_appstate import *
from kanren_log import *
import kanren_settings
from kanren_dict import *
from kanren_showarea import *
from kanren_useroracle import *
from kanren_util import *
from kanren_varlabel import *
from kanren_wordadder import *

STATUS_LINE_LENGTH=10
INFO_FIELD_LENGTH=10

INTERNAL_DICT_SPECIFIER='.' # For debugging: This string denotes the "internal dictionary" holding the testdata

# Main class, representing the KanRen application
class KanRen(object):

    # _root_window: Tk root window object (of class Tk)
    # _settings: Object of type kanren_settings.Settings, containing the settings for the application
    # _initial_dictionary: None, or a path to a file holding a dictionary to load, or the special
    #     constant INTERNAL_DICT_SPECIFIER, which loads the statically defined test dictionary. This
    #     is used for debugging only.
    # _debug_mode: If True, additional buttons and fields are shown to aid in debugging.
    def __init__(self,_root_window,_settings,_initial_dictionary,_debug_mode,_initial_action=None):
        set_debug_mode_to(_debug_mode)
        self.rootw=_root_window
        self.S=_settings
        self.D=GoiDict(_settings) # Initially empty dictionary
        top_chooser=Frame(self.rootw) # Buttons for loading/saving a dictionary, starting the quiz, etc.
        # Top row of buttons
        top_chooser.pack(side=TOP)
        if _debug_mode:
            button_debug=Button(top_chooser,text='状態',command=self.show_state)
            button_debug.pack(side=RIGHT)
        self.button_quit=Button(top_chooser,text=self.S.quit_button_text(),command=self.quit)
        self.button_quit.pack(side=RIGHT)
        self.button_save_dict=Button(top_chooser,text=self.S.save_dict_button_text(),state=DISABLED,command=self.save_dict)
        self.button_save_dict.pack(side=RIGHT)
        self.button_save_dict_as=Button(top_chooser,text=self.S.save_dict_as_button_text(),command=self.save_dict_as)
        self.button_save_dict_as.pack(side=RIGHT)
        self.button_load_dict=Button(top_chooser,text=self.S.load_dict_button_text(),command=self.load_dict)
        self.button_load_dict.pack(side=LEFT)
        self.button_merge_dict=Button(top_chooser,text=self.S.merge_dict_button_text(),command=self.merge_dict)
        self.button_merge_dict.pack(side=LEFT)
        self.button_add_word=Button(top_chooser,text=self.S.add_word_button_text(),command=self.add_word)
        self.button_add_word.pack(side=RIGHT)
        self.button_start=new_repr_list()
        for i in ALLREP_RANGE:
            self.button_start[i]=Button(top_chooser,text=self.S.start_button_text(i),state=DISABLED,command=lambda num=i: self.start(num))
            self.button_start[i].pack(side=LEFT)            
        # The "Showarea" is the area where the words are shown in the different representations    
        # self.showarea=Label(self.rootw,text="test test 始めましょう", anchor=NE, borderwidth=6, relief=GROOVE)
        self.showarea=new_repr_list()
        initial_text=new_repr_list()
        initial_text[KANJI]="頑張りましょう"
        initial_text[KANA]="がんばりましょう"
        initial_text[GAIKOKUJI]="✌"
        for i in ALLREP_RANGE:
            self.showarea[i]=Showarea(self.rootw,self.S,initial_text[i],i,self.D)
            self.showarea[i].pack()
        # Below the showrea, we have a message line, a button for requesting the next challenge, and
        # a status area.
        bottom_controls=Frame(self.rootw)
        bottom_controls.pack()
        self.status_line=VarLabel(bottom_controls,"no dict",borderwidth=0,relief=SUNKEN,width=STATUS_LINE_LENGTH)
        self.info_field=VarLabel(bottom_controls,"",borderwidth=0,relief=GROOVE,width=INFO_FIELD_LENGTH)
        self.info_field.pack(side=RIGHT,padx=9)
        self.status_line.pack(side=RIGHT,padx=10) # Without padx, status_line would stick next to button_next_challenge
        # Initially, a completely new challenge is shown (for example, just the Kanji version of the
        # word). Every subsequent click on this button, it fills in one by one the other kinds of 
        # the word. If all kinds in the showarea are filled in, the user is asked, whether the
        # challenge should be considere resolved correctly, and the cycle begins with the next
        # challenge.
        self.button_next_challenge=Button(bottom_controls,text=self.S.next_challenge_text(),state=DISABLED,command=self.next_challenge)
        self.button_next_challenge.pack(side=RIGHT)
        self.messageline=VarLabel(bottom_controls," "*24,borderwidth=0,relief=FLAT,width="80")
        self.messageline.pack(side=RIGHT)
        self.oracle=Oracle(self.rootw,self.S,self) # User oracle (initially invisible)
        self.event_history_number=0    
        self.wordadder=Wordadder(self.rootw,self.S,self.D,self.last_event_number) # Button to add a new word (initially invisible)
        self.current_kind=UNDEFINED_KIND
        if _initial_dictionary:
            # Dictionary should be loaded (passed via command line)
            if _initial_dictionary == INTERNAL_DICT_SPECIFIER:
                self.D.load_testdata()
                self.rootw.wm_title('Internal dictionary used for debugging')
                self.enable_start_buttons()
                self.update_status_line()
            else:
                self.load_dict_by_name(_initial_dictionary)
        else:
            self.rootw.wm_title(self.D.current_dict_pp()) # should display "no dictionary"
        if _initial_action is not None:
            self.button_start[_initial_action].invoke()
            
    def show_state(self):
        self.D.pp()
        
    def enable_save_button(self):
        if self.D.dictionary_pathname():
            self.button_save_dict['state']=NORMAL    
        
    def set_start_buttons_state(self, new_state):
        for i in ALLREP_RANGE:
            self.button_start[i]['state']=new_state
    
    def enable_start_buttons(self):
        self.set_start_buttons_state(NORMAL)
        self.enable_save_button()

    def disable_start_buttons(self):
        self.set_start_buttons_state(DISABLED)
        self.button_save_dict['state']=DISABLED
            
    def load_dict_by_name(self,pathname,merge=False):
          try:
            self.D.dict_import(pathname,merge)
          except Exception:
            e=sys.exc_info()[1]
            self.messageline.settext("Oh je, oh je")
            showerror("Import Error",str(e)) 
            self.messageline.settext(self.S.msg__import_error())
          else:
            if merge:
                self.messageline.settext(self.S.msg__dict_merged())
            else:
                self.event_history_number=0    
                self.messageline.settext(self.S.msg__dict_loaded())
          self.update_status_line()
          if self.D.n_entries() > self.S.do_not_ask_again_steps():
            self.enable_start_buttons()
          else:
            self.disable_start_buttons()
          self.rootw.wm_title(self.D.current_dict_pp())
    
    def load_dict(self):
        # TODO: A dirty dictionary should give three options, not two (save and proceed, do not save
        # and proceed, cancel). Until this is solved, we don't use the settings for the strings to
        # be displayed.
        if not (self.D.is_dirty() and not askokcancel("Dictionary dirty","本当ですか")):
            self.load_dict_unconditionally()
            
    def load_dict_unconditionally(self,merge=False):
        self.clear_message_and_info()
        log("Loading dictionary, merge=",merge)
        pathname=ucguard(askopenfilename(initialdir='.',title=self.S.choose_dict_window_title()))
        if pathname:
            self.load_dict_by_name(pathname,merge)
        return pathname
        
    def merge_dict(self):
        self.load_dict_unconditionally(True)
        
    # Start training    
    def start(self,kind):
        self.clear_message_and_info()
        if self.current_kind != kind and self.current_kind != UNDEFINED_KIND:
            self.end_challenge()
        if self.D.n_entries() == 0:
            self.load_dict()
        if self.D.n_entries() == 0:
            raise Exception("Dictionary empty")
        if self.D.n_entries() <= self.S.do_not_ask_again_steps():
            raise Exception("Dictionary too small. Must have at least "+str(1+self.S.do_not_ask_again_steps())+" entries")
        self.current_kind=kind # For this training, "kind" is displayed as the first challenge
        self.current_sequence=SEQUENCE_FOR[kind] # These will be subsequently displayed, whenever
                                                 # the user clicks on button_next_challenge
        if self.D.n_entries_for_kind(kind) == 0:
            raise Exception("Dictionary empty for kind "+str(kind)+" ("+KIND_NAME[kind]+")")
        self.prepare_new_cycle()
        self.next_challenge()  
        return self
    
    def prepare_new_cycle(self):
        self.oracle.unshow()
        # self.button_next_challenge['state']="normal"
        self.event_history_number=self.event_history_number+1
        show_widget_enabled(self.button_next_challenge)
        log("current kind:",self.current_kind,"event:",self.event_history_number,"dictionary dirty",self.D.is_dirty())
        self.choose_goi=self.D.choose(self.current_kind,self.event_history_number)
        log(KIND_NAME[self.current_kind],"chosen:",self.choose_goi.pp())
        self.choose_goi.set_event_number(self.event_history_number)
        for i in ALLREP_RANGE:
            self.showarea[i].settext('?')
            self.showarea[i].disable_edit()
        self.current_cycle_yet_to_be_done=self.current_sequence[:]
        self.current_cycle_yet_to_be_done.insert(0,self.current_kind)
        log("Cycle",self.current_cycle_yet_to_be_done)
        self.write_info_field(str(self.event_history_number)+'.',self.choose_goi.kotoba(self.current_kind).level_count_pp())
    
    # Present the next challenge
    def next_challenge(self):
        # Pop first element from the list of kinds to be shown
        display_kind=self.current_cycle_yet_to_be_done.pop(0)
        self.showarea[display_kind].settext(self.choose_goi.gettext(display_kind))
        self.showarea[display_kind].enable_edit()
        len_ccytbd=len(self.current_cycle_yet_to_be_done)
        if len_ccytbd == REPRESENTATIONS-2:
            self.messageline.clear()
        if len_ccytbd == 0:
            # We have now shown the word of the current cycle in all kinds. Now the user needs to
            # evaluate the challenge.
            show_widget_disabled(self.button_next_challenge)
            # Asking the user for his evaluation
            self.oracle.show()

    # See, whether the user has applied some editing on the words in the challenge. This includes
    # removing a Kanji definition completely. The dictionary is adjusted to reflect the change.
    def process_edited_words(self):
        bark()
        for kind in ALLREP_RANGE:
            text=self.showarea[kind].gettext()
            if self.showarea[kind].has_unsaved_edits():
                if text == '':
                    user_prompt=self.S.fmtmsg__showarea_has_deleted_text() % KIND_NAME[kind]
                else:
                    user_prompt=self.S.fmtmsg__showarea_has_unsaved_changes() % text
                if askyesno("Saving Changes",user_prompt):
                    self.showarea[kind].save_edit()
                else:
                    self.showarea[kind].cancel_edit(True) # No need to ask again about discarding changes
            if self.showarea[kind].is_dirty():
                if text == '': # Showarea (via kanren_dict) ensures that this can happen only if the kind is OPTIONAL
                    self.D.remove_optional_representation(self.choose_goi)
                else:
                    self.D.update_word(kind,self.choose_goi,text)
                    # Plausibility check
                    if not self.D.already_exists(kind,text):
                        raise "Dictionary update failed for "+text
                self.showarea[kind].set_dirty_to(False)
                self.D.set_dirty()
        self.update_status_line()
    
    # Callback from User Oracle. This will be invoked, after the user judged, whether he has
    # answered correctly.
    def continue_from_oracle(self,_answer_correct_p):
        log(_answer_correct_p)
        self.process_edited_words(); # See whether the user has edited some definitions on the fly
        changed_level=self.D.answered_correctly(self.choose_goi,self.current_kind,_answer_correct_p)
        if changed_level is not None:
            self.messageline.settext(self.S.fmtmsg__level_changed_to() % changed_level)
        # self.status_line.settext('BBBB')
        # self.info_field.settext('CCCC')
        # Should we mark the dictionary as dirty after the cycle? In general, something will have
        # changed (hit count, level), but this means that the dictionary will be dirty virtually
        # all the time. Therefore, we consider it dirty only if something has changed at the
        # vocabulary itself
        # self.D.set_dirty()
        self.prepare_new_cycle()
        self.next_challenge()
    
    def end_challenge(self):
        self.current_kind=UNDEFINED_KIND
        
    def add_word(self):
        self.wordadder.show()
        
    def add_word_destroy_callback(self):
        bark()
        self.wordadder=None
        self.update_status_line()
        
    def last_event_number(self):
        return self.event_history_number
        
    def save_dict(self,pathname=''):
        if self.D.dict_export(pathname):
            self.D.set_clean()
            if pathname == '':
                add_info=''
            else:
                add_info=' ('+basename(pathname)+')'
            if self.D.n_entries() > self.S.do_not_ask_again_steps():
              self.enable_start_buttons()
            self.messageline.settext(self.S.msg__dict_saved()+add_info)
            self.update_status_line()
        else:
            self.messageline.settext(self.S.msg__dict_not_saved())
            
    def save_dict_as(self):
        # Python 2: Make sure that pathname ends up as str, not as type unicode (doesn't seem to
        #           occur on the Mac, but with Python 2.7 on the PC)
        pathname=ucguard(asksaveasfilename(initialdir='.',title=self.S.choose_dict_window_title()))
        if pathname:
            self.save_dict(pathname)
            self.enable_save_button()
                        
    def update_status_line(self):
        dirty_indicator=' '
        if self.D.is_dirty():
            dirty_indicator='*'
        self.status_line.settext(dirty_indicator+' '+str(self.D.n_entries()))
        
    def write_info_field(self,*_data):
        bark()
        self.info_field.settext(" ".join(map(str,_data)))
        
    def clear_message_and_info(self):
        bark()
        self.messageline.clear()
        self.info_field.clear()
        
    def quit(self):
        self.rootw.quit()
        
def usage():
    me=basename(sys.argv[0])
    prnt(me,"[-l |--load=]DICTIONARY [--debug] [--trace=FILENAME] [--dnamin]")
    prnt(me,"[-? | --help]")
    prnt('''
    --load=DICTIONARY[~ACTION] : Initially load DICTIONARY. Optionally start immediately an ACTION:
                                 ACTION = 0 : Start querying KANJI
                                 ACTION = 1 : Start querying KANA
                                 ACTION = 2 : Start querying in foreign language
                                 Example: --load=mydict.txt~2
    
    DEBUGGING AIDS:
    --debug : Decorate GUI with debugging aids
    --trace=FILENAME : Debug trace goes to FILENAME instead of stdout
    --dnamin : Set "do-no-ask-again-steps" to a minimal value
    ''')
    
#------------------ MAIN ---------------------------

initial_dictionary=''
initial_action=None
debug_mode=False
dnamin_override=None
try:
    options,remaining = getopt(sys.argv[1:],'dl:t:?',['load=','debug','trace=','dnamin','help'])
    if len(remaining) > 0:
        raise Exception("Unrecognized arguments: "+str(remaining))
    for o,a in options:
        if o in ('-l','--load'):
            a_fields=a.split('~')
            if(len(a_fields)>2):
                raise Exception("Restriction: Dictionary file name must not contain a semicolon (ambiguous)")
            dict_to_load=a_fields[0]
            if(len(a_fields) == 2):
                initial_action=a_fields[1]
            if(not (initial_action in [None,'0','1','2'])): # TODO replace by mapping
                raise Exception("Unrecognized action: "+initial_action)
            if initial_action is not None:
                initial_action=int(initial_action)
            if initial_dictionary:
                raise Exception("Initial dictionary specified twice")
            if dict_to_load != INTERNAL_DICT_SPECIFIER and not os.path.isfile(dict_to_load):
                raise Exception("Dictionary file "+dict_to_load+" does not exist")
            initial_dictionary=dict_to_load
        if o == '--dnamin':
            dnamin_override=0
        if o in ('-d','--debug'):
            debug_mode=True
        if o in ('-t','--trace'):
            set_default_debug_file(a)
        if o in ('-?','--help'):
            usage()
            sys.exit(0) # This throws SystemExit
except SystemExit:
    raise
except Exception:
    e=sys.exc_info()[1]
    prnt(e)
    usage()
    sys.exit(1)
    
root=Tk()
screenwidth=root.winfo_screenwidth()
screenheight=root.winfo_screenheight()
x=screenwidth/12
y=screenheight/32
#w=screenwidth*5/6
#h=screenheight/4
#root.geometry("%dx%d+%d+%d" % (w,h,x, y))
root.geometry("+%d+%d" % (x, y))
s=kanren_settings.settings()
if not dnamin_override is None:
    s.set_do_not_ask_again_steps(dnamin_override)
kanrenapp=KanRen(root,s,initial_dictionary,debug_mode,initial_action)
root.mainloop()
log("Mainloop ended")
root.destroy()
