# -*- coding: utf-8 -*-

"""Implements the Japanese-Foreign dictionary for kanren"""

import math
import os.path
import re

# from sets import * # Deprecated since 2.6
try:
    set # Will raise exception in 2.4 or later
except NameError:
    from sets import Set as set, ImmutableSet as frozenset # Set is called set now, because builtin 
         # type. ImmutableSet becomes builtin frozenset.

from shutil import copyfile
try:
  from tkMessageBox import showerror,showwarning # TODO: Maybe we shouldn't let this module depend on Tk?
except ImportError:
  from tkinter.messagebox import showerror,showwarning
from random import randint, choice # Don't use whrandom (deprecated in 2.4)

from comment_file_input import *
from kanren_log import *
from kanren_util import bool_pp,prnt,uopen

DICT_VERSION = 'KANREN01'

UNDEFINED_KIND = -1
KANJI = 0
KANA = 1
GAIKOKUJI = 2
OPTIONAL_KIND=KANJI

REPRESENTATION_MISSING_STRING = '-' # Denotes in the file representation, that this
    # word has no representation in this kind. Currently makes only sense if kind 
    # is KANJI
REPRESENTAION_FORBIDDEN_CHARACTERS = r'#\@$' # The user must not use these 
    # characters in a representation, because they are reserved for use in the
    # dictionary
REPRESENTATION_FORBIDDEN_CHARACTERSET = frozenset(REPRESENTAION_FORBIDDEN_CHARACTERS)
REPRESENTATION_FORBIDDEN_BORDER_CHARACTERS = r':;+-/&=<>' # The user must not use 
    # these characters at the first or last character in a representation
REPRESENTATION_FORBIDDEN_BORDER_CHARACTERSET = frozenset(REPRESENTATION_FORBIDDEN_BORDER_CHARACTERS)

def contains_forbidden_character(text):
    """Returns true iff text contains a character which is not permitted in a dictionary word. These
       are special characters which have a particular meaning during dictionary import.
    """
    return len(set(text).intersection(REPRESENTATION_FORBIDDEN_CHARACTERSET)) > 0
    
def has_contains_forbidden_character_at_border(text):
    return len(text) > 0 and len(set(text[0]+text[-1]).intersection(REPRESENTATION_FORBIDDEN_BORDER_CHARACTERSET)) > 0
    
REPRESENTATIONS = 3 # KANJI, KANA, GAIKOKUJI

ALLREP_RANGE = range(0,REPRESENTATIONS)

KIND_NAME = {
    UNDEFINED_KIND : 'UNDEFINED',
    KANJI : '漢字',
    KANA  : 'かな',
    GAIKOKUJI : 'ANDERS'
}

# Sequence of kinds to be presented. For example, if the first challenge is the word in Kanji,
# the user must first read it (i.e. produce the Kana version), and finally understand it.

SEQUENCE_FOR = [
    [KANA,GAIKOKUJI],  # KANJI
    [GAIKOKUJI,KANJI], # KANA
    [KANA,KANJI]       # GAIKOKUJI
]    

# Returns a fresh list for storing information about the representations
def new_repr_list():
    return [None]*REPRESENTATIONS
    
# Normalize a word by removing non-needed white space
def whitespace_normalize_word(word):
    # Also replaces runs of whitespaces by a single space
    return ' '.join(word.strip().split())

# Level 0 is not used yet (maybe later for deposing words which should not be queried at the moment)    
# Level 1 are the easiest words, level max_level are the most difficult ones.
# Each word has associated a hit count (atari), which is initially set to 0.
# If the word is known, the hit count is incremented, otherwise decremented.
# Configurable parameters and their defaults: X=3, Y=3, F=25
# If the hit count reaches X (X>0):
#   If the word is at level n with n>1, it goes to level n-1 and the hit count is set to 0.
#   If the word is at level 1, the hit count stops incrementing.
# If the hit count reaches -Y (Y>0):
#   If the word is at level max_level, and the number of words at max_level is less than or equal to
#     max(3,T/F), where T is the total number of word in the dictionary, the hit count stops
#     decrementing.
#   Otherwise, the word goes to the next higher level (extending max_level if necessary) and the
#     hit count is set to 0.

# The user should not be asked the same word in succession, because this would not have a learning
# affect. Each time the user is asked, a history number is increased, and this history number
# is stored with each word transiently (not saved to database). Only those words are elegible, 
# where the difference between the current history number and their history number is less than
# a certain (usually low), configurable number.
# The history number is associated with each triple (KANJI,KANA,GAIKOKUJI), i.e. stored in the
# objects of type Goi, and not with the individual words (type Kotoba).

# Random chooser:
# Level        1    2    3    4    5
# One of 1 22 333 4444 55555
# Probability level N is 2/(N+1)
# Using Conditional Probabilities
# If max_level == M, we have T=M*(M+1)/2 token.
# Getting a token at level 1 has probability 1/T = U0/V0
# at level 2, if level 1 has been excluded has probability 2/14 = (U0+1)/(V0-U0) = U1/V1
# at level 3 : 3/12 = (U1+1)/(V1-U1) = U2/V2
# etc. until U==V
def random_level(max_level):
    u=1
    v=max_level*(max_level+1)/2
    while u != v:
        if randint(1,v) <= u:
            break
        v=v-u    
        u=u+1
    return u
    
# Word to be presented to the user, level of difficulty, and hit cound (increases for each correct
# answer and decreases for each incorrect answer).  
class Kotoba(object):
    """One word, consisting of textual representation, difficulty level and hit count"""
    def __init__(self,_text,_initial_level,_atari=0):
        self.txt=_text   # Textual representation, for instance "Traube" or "葡萄" or "ぶどう"
        self.lvl=_initial_level # Difficulty level for this word
        self.atari=_atari # Initial hit count for this word
        
    # Textual representation. Don't use builtin repr(), because this would turn the japanese 
    # characters into their hex escape sequences.
    def text(self):
        return self.txt
        
    # Set new textual representation.    
    def settext(self,_newtext):
        self.txt=_newtext
        return self
        
    # Level of difficulty  
    def level(self):
        return self.lvl
    
    # Set new difficulty level
    def set_level(self,_new_level):
        self.lvl=_new_level
        return self
        
    # Hit count (incremented for correct answer, decremented for wrong answer)   
    def hit_count(self):
        return self.atari
        
    # Nicely format level and hit count    
    def level_count_pp(self):
        return str(self.lvl)+','+str(self.atari)
        
    # Pretty-print for i.e. debug output    
    def pp(self):
        return '/'+self.txt+'/{'+self.level_count_pp()+'}'
    
    def adjust_hit_count(self,_settings,_answer_correct_p,_max_allowed_level):
        """Adjusts the hit count. If the level shifts to a new one, returns the OLD level. Otherwise
        returns None. Note: The level is NOT changed here. This needs to be done by the caller.
        """
        old_level=self.lvl
        new_level=old_level
        old_atari=self.atari
        bark("Adjusting",self.txt,"from level",old_level,"/ old hit count",self.atari)
        if _answer_correct_p:
            # Increment hit count and shift down if necessary
            self.atari=self.atari+1
            if self.atari >= _settings.dict_atari_X():
                if old_level > 1:
                    new_level=old_level-1
                    self.atari=0
                else:
                    log("Not adjusting level, because already at lowest level")
                    self.atari=old_atari
        else: # answer was not correct
            # Decrement hit count and shift up if necessary
            self.atari=self.atari-1
            if self.atari <= -_settings.dict_atari_Y():
                if old_level < _max_allowed_level:
                    new_level=old_level+1
                    self.atari=0
                else:
                    log("Not adjusting level, because already at maximum allowed level",_max_allowed_level)
                    self.atari=old_atari
        log("Object of type Kotoba adjusted hit count from",old_atari,'to',self.atari)
        if old_level != new_level:
            log("Level is supposed to change from",old_level,'to',new_level)
            return old_level
        else:
            return None

# Word in 3 Forms (Kanji, Kana, foreign language). The Kanji version can be the null string. 
# By default, all 3 kinds are assigned to the same level (_initial_level), but if _initial_level is
# an array of 3 numbers, _initial_level[kind] is used as the level for 'kind'
class Goi(object):
    def __init__(self,_initial_last_asked_event_number,_kanji,_kana,_gaikoku,_initial_level,_kanji_atari=0,_kana_atari=0,_gaikoku_atari=0):
        self.ji=new_repr_list()
        if isinstance(_initial_level,list): # OR: type(_initial_level) == type(list()):
            initial_levels=_initial_level
        else:
            initial_levels=[_initial_level]*REPRESENTATIONS
        self._make_kotoba_helper(KANJI,_kanji,initial_levels,_kanji_atari)
        self._make_kotoba_helper(KANA,_kana,initial_levels,_kana_atari)
        self._make_kotoba_helper(GAIKOKUJI,_gaikoku,initial_levels,_gaikoku_atari)
        self.most_recently_queried=_initial_last_asked_event_number

    # Helper function for constructor. 
    def _make_kotoba_helper(self,_kind,_text,_initial_levels,_atari):
        if len(_text) > 0:
            self.ji[_kind]=Kotoba(_text,_initial_levels[_kind],_atari)
        elif _kind != OPTIONAL_KIND:
            raise Exception("Plausibility check violated: empty "+KIND_NAME[_kind])
       
    # Remove textual representation for the kind supplied. Note that this is done even if the
    # kind is NOT optional.
    def remove(self,_kind):
        self.ji[_kind]=None
            
    # Returns the Kotoba object of the supplied kind
    def kotoba(self,_kind):
        return self.ji[_kind]
        
    # Explicitly resets the event number to the supplied value. Note that this number is initially None.    
    def set_event_number(self,_event_number):
        self.most_recently_queried=_event_number
        return self
    
    # Returns event number
    def get_event_number(self):
        return self.most_recently_queried
    
    # Returns the textual representation of the word for a certain kind. If there is none, because
    # we have an optional kind where there is no kotoba object is associated, the null string is
    # returned.
    def gettext(self,_kind):
        kotoba_for_type=self.kotoba(_kind)
        if kotoba_for_type is None:
            return ''
        else: 
            return kotoba_for_type.text()
            
    # Provide a new representation for a text.
    #  _kind    : Kind of text (for instance: KANJI)
    #  _newtext : The new text
    # Returns: The Kotoba object associated to _kind
    # Side effect: If this is a text, where there has been no text before (for example, a Goi with 
    #              no Kanji specification), a new Kotoba object is implicitly created.
    def settext(self,_kind,_newtext):
        kotoba_for_type=self.kotoba(_kind)
        if kotoba_for_type is None:
            bark("Create new Kotoba object for this kind (probably, a Kanji was added where there had been none before)")
            kotoba_for_type=self.ji[_kind]=Kotoba(_newtext,1)
        else:
            kotoba_for_type.settext(_newtext)
        return kotoba_for_type
            
    def adjust_hit_count(self,_settings,_kind,_answer_correct_p,_max_allowed_level):
        """Does an adjust_hit_count for the Kotoba object corresponding to _kind. Return value as in
        Kotoba.adjust_hit_count(...)
        """
        return self.ji[_kind].adjust_hit_count(_settings,_answer_correct_p,_max_allowed_level)
    
    def pp(self):
        """Returns a string containing a (more or less) prettily formatted representation of this
        object. Used mainly for debugging.
        """
        rep=""
        for i in ALLREP_RANGE:
            rep=rep+KIND_NAME[i]+':'
            if(self.ji[i]):
                rep=rep+self.ji[i].pp()
            rep=rep+'. '
        return rep
                
# This class organizes all triples from the viewpoint of a certain kind (for instance: KANA).
# The difficulty level is organized for this kind.
# An AtLevel object for kind KANA for instance sorts the triples according to their KANA difficulty
# levels.
# Note that the KIND itself doesn't need to be stored in this class. It is sufficient, if the
# user of this class knows it, because this information is implicitly passed via the add() function.

class AtLevel(object):        

    # member variables:
    # from_at_level[n] holds all triples which are at difficulty level n for this kind.
    def __init__(self,_kind,_settings):
        # Level 0 is not used at the moment, but is reserved for cemetery feature
        # (characters which won't be queried, but still are stored in the dictionary).
        # For a level L, self.from_at_level[L] is a dictionary, mapping a word (text) to its
        # Goi object.
        self.from_at_level=[{}] # Level 0 only
        self.max_level=0
        self.goi_pool={} # Goi dictionariy, with the text of the respective kind as a key
        self.kind=_kind
        self.S=_settings
        
    def add(self,kotoba,goi):
        if kotoba is not None:
            log(kotoba.pp())
            # Make sure that the word does not exist yet. Example:
            # (暑い,あつい,heiß) and (厚い,あつい,dick) would not work if we ask for あつい
            # This must be disambiguated by the user to, i.e.
            # (暑い,あつい[1],heiß (Wetter)) and (厚い,あつい[2],dick (Buch))
            # or
            # (今日は厚い,きょうはあつい,heute ist es heiß) etc.,
            kotoba_text=kotoba.text()    
            if kotoba_text in self.goi_pool:
                raise Exception(self.S.msg__entry_already_exists()+kotoba_text)
            if kotoba.level() > self.max_level:
                # Note: range(3,6) is (3,4,5), not (3,4,5,6)
                for i in range(self.max_level+1,kotoba.level()+1):
                    self.from_at_level.append({})
                log("Adjusting max_level from",self.max_level,"to",kotoba.level())
                self.max_level=kotoba.level()
            self.from_at_level[kotoba.level()][kotoba_text]=goi
            self.goi_pool[kotoba_text]=kotoba
            # Plausibility check
            if not self.already_exists(kotoba_text):
                raise Exception(kotoba+" not in pool") 
                
    def update_pool(self,old_text,new_text):
        self.goi_pool[new_text]=self.goi_pool[old_text]
        del self.goi_pool[old_text]
        return self

    # The representation for this goi must be removed. The caller is responsible to ensure that this
    # operation is permitted.
    def remove_representation(self,goi):
        if self.kind != OPTIONAL_KIND:
            raise Exception("Plausibility check failed. remove_representation invoked for non-optional kind")
        old_text=goi.gettext(self.kind)
        log("Removing representation:",old_text)
        del self.goi_pool[old_text]
        kotoba_level=goi.kotoba(self.kind).level()
        del self.from_at_level[kotoba_level][old_text]
        if len(self.from_at_level[kotoba_level]) == 0:
            self.level_compactification(kotoba_level)
        goi.remove(self.kind)
        return self
        
    def update(self,goi,new_text):
        old_text=goi.gettext(self.kind)
        # This creates and returns a new kotoba object if no definition for this kind was present:
        kotoba=goi.settext(self.kind,new_text) 
        if old_text:
            self.update_pool(old_text,new_text)
            kotoba_level=goi.kotoba(self.kind).level()
            # Plausibility check
            if not (old_text in self.from_at_level[kotoba_level]):
                raise "Plausibility check failed: '"+old_text+"' should be at level "+str(kotoba_level)+" for kind "+KIND_NAME[self.kind]
            self.from_at_level[kotoba_level][new_text]=self.from_at_level[kotoba_level][old_text]
            del self.from_at_level[kotoba_level][old_text]
        else: # A new definition has been provided (this must be an OPTIONAL_KIND)
            # The kotoba object has been created with a new default level. This will likely be 1,
            # because the Goi object doesn't know anything about the other words in the pool. We
            # will raise it to the level of maximum difficulty, because it is new.
            kotoba_level=max(kotoba.level(),self.maxlevel())
            kotoba.set_level(kotoba_level)
            # Border case: It could be, that from_at_level doesn't have any entry yet for this
            # level. This happens if we start with a dictionary which doesn't have any Kanji 
            # definitions yet.
            while(len(self.from_at_level)<=kotoba_level):
                self.from_at_level.append({})
                self.max_level=self.max_level+1
            self.from_at_level[kotoba_level][new_text]=goi
            self.goi_pool[new_text]=kotoba
        return self
    
    def n_entries(self):
        n=0
        for i in range(0,self.max_level+1):
            n=n+len(self.from_at_level[i])
        return n
            
    def n_entries_at_max_level(self):
        return len(self.from_at_level[self.max_level])
        
    def is_at_level(self,text):
        if text in self.goi_pool:
            return self.goi_pool[text].level()
        else:
            return -1

    def already_exists(self,text):
        return self.is_at_level(text) >= 0
        
    # The maximum level a word is allowed to be placed. Usually, this is one more than the maximum
    # level existing so far, unless there are "only few" words in this level, in which case we don't
    # allow the levels to be increased.
    def maximum_allowed_level(self,_settings):
        result=self.max_level
        n_entries_here=self.n_entries_at_max_level()
        limiting_parameter=3 # TODO: Should this be configurable?
        if(n_entries_here > max(limiting_parameter,int(math.ceil(self.n_entries()/float(_settings.dict_atari_F()))))):
            result=result+1
        log(result,'(max_level for',KIND_NAME[self.kind],'is',str(self.max_level)+', number of entries here:',str(n_entries_here)+')')
        return result
            
    def translations(self,_level):
        return self.from_at_level[_level]

    # Return: suitable random word at specific level, or None if nothing found         
    def non_recent_random_word_at_level(self,_level,_event_history_number,_do_not_ask_again_steps):
        words=list(self.from_at_level[_level].keys())
        word=None
        log(_level,_event_history_number,_do_not_ask_again_steps)
        while word is None and len(words) > 0:
            word=choice(words)
            log("trying",word,"with recent history number",self.from_at_level[_level][word].get_event_number(),"from",len(words),"word(s)")
            if _event_history_number - self.from_at_level[_level][word].get_event_number() <= _do_not_ask_again_steps:
                log("rejected, because too recent")
                words.remove(word)  # By design, word must exist exactly once in words.
                word=None
        return word # word as a string, or None if no suitable word has been found.
        
    # Return: suitable random word as object of type Goi
    def choose(self,_event_history_number,_do_not_ask_again_steps):
        level=random_level(self.max_level)
        log("try first to choose from level",level)
        word=self.non_recent_random_word_at_level(level,_event_history_number,_do_not_ask_again_steps)
        if word is None:
            # No suitable words found at the randomly chosen level. This should happen rarely, only
            # for levels with very few words, where all of them have been asked already recently. In
            # this case, we just go through all the other levels in sequence.
            for l in range(self.max_level,0,-1): # Loop goes from max_level to 1
                if l != level: # We can skip the level which we already have tried
                    word=self.non_recent_random_word_at_level(l,_event_history_number,_do_not_ask_again_steps)
                    if word is not None:
                        level=l
                        log("Using replacement level",level)
                        break
                    log("Nothing found at level",l)
        if word is None: # No non-recent word found on any level. 
            if _do_not_ask_again_steps == 0:
                msg='<'+self.S.msg__no_suitable_word()+'>'
                return Goi(0,msg,msg,msg,1)
            else:
                return self.choose(_event_history_number,0)
        return self.from_at_level[level][word] # Goi
    
    # Invoke this function after an operation, which possibly could make a level empty. This could
    # be the case because a word has been answered correctly, or incorrectly, to often, or because
    # a Kanji definition has been deleted.
    # _level_to_test : The level which might be empty now, or None. If it is None, compactifies
    #                  for all levels. This can be used after loading a dictionary, which may
    #                  contain empty level definitions.
    def level_compactification(self,_level_to_test=None):
        compactification_performed=False
        if _level_to_test is None:
            # We can not loop over self.level_range(), because due to interim compactification,
            # the level_range could shrink during the loop. Hence, an explicit 'while' loop is used
            level_to_test=1
            while level_to_test <= self.maxlevel():
                compactification_performed_on_this_level = self.level_compactification(level_to_test)
                compactification_performed = compactification_performed or compactification_performed_on_this_level
                if not compactification_performed_on_this_level:
                    level_to_test=level_to_test+1
        elif len(self.from_at_level[_level_to_test].keys()) == 0:
            log("Removing level",_level_to_test,"for",KIND_NAME[self.kind],"because it became empty")
            for shift_to_level in range(_level_to_test,self.max_level):
                log("Shifting from level",shift_to_level+1,"to level",shift_to_level)
                self.from_at_level[shift_to_level]=self.from_at_level[shift_to_level+1]
                for goi in self.from_at_level[shift_to_level].values():
                    goi.kotoba(self.kind).set_level(shift_to_level)
            # Drop highest level (which now became empty)
            del self.from_at_level[-1]
            self.max_level=self.max_level-1 
            compactification_performed=True
        return compactification_performed
        
    def shift_level(self,_goi,_from_level,_to_level):
        kotoba_obj=_goi.kotoba(self.kind)
        key=kotoba_obj.text()
        log("shift level for",key,'from',_from_level,'to',_to_level)
        # Plausibility checks:
        # If _to_level equals 1, nothing needs to be done (but the caller should ensure that this doesn't happen)
        # if _to_level is one higher than maxlevel, create a new level
        # If after removing the element from _from_level, that level becomes empty, we need to 
        #    contract everything (i.e. shift all higher elements down and adjust their level)
        if abs(_from_level-_to_level) != 1:
            raise Exception("Source and target level do not differ by 1")
        if _to_level <= 0:
            raise Exception("Plausibility violation: level should be shifted to non-positive value")
        if _to_level > self.max_level:
            self.from_at_level.append({})
            self.max_level=self.max_level+1
        # Remove goi from _from_level
        del(self.from_at_level[_from_level][key])
        # Change the kotoba of type self.kind to _to_level
        kotoba_obj.set_level(_to_level)
        # Add goi to _to_level
        self.from_at_level[_to_level][key]=_goi
        self.level_compactification(_from_level) # Level might be empty now
        return self
        
    def maxlevel(self):
        return self.max_level
        
    def level_range(self):
        return range(1,self.max_level+1)
        
    def each_goi(self):
        for level in self.level_range():
            for key in self.from_at_level[level].keys():
                yield self.from_at_level[level][key] # Object of type Goi
        
class GoiDict(object):

    # member variables:
    # nentries : Number of entries (triples) in the dictionary
    # from_ji_type[kind] : Instance of type AtLevel, which holds all entries of this kind.
    
    def __init__(self,_settings):
        self.S=_settings
        self.reset()
        
    def reset(self):
        self.initial_last_asked_event_number=-self.S.do_not_ask_again_steps()
        self.dirty=False # TODO: Set to true if dictionary has been edited (changes, extensions, deletions) since last save
        self.from_ji_type=new_repr_list()
        for i in ALLREP_RANGE:
            self.from_ji_type[i]=AtLevel(i,self.S)
        self.nentries=0
        self.current_dict_pathname='' # Path from where the current dictionary has been read
        
    def current_dict_pp(self):
        return self.current_dict_pathname or 'No Dictionary'
        
    def dictionary_pathname(self):
        return self
        
    def already_exists(self,kind,word_text):
        return self.from_ji_type[kind].already_exists(word_text)
        
    def already_has(self,goi):
        result=True
        for i in ALLREP_RANGE:
            if goi.kotoba(i) is not None:
                result = result and self.already_exists(i,goi.kotoba(i).text())
        return result

    def add(self,goi):
        if not self.already_has(goi):
          for i in ALLREP_RANGE:
            if goi.kotoba(i) is not None:
                word_text=goi.kotoba(i).text()
                if self.already_exists(i,word_text):
                    raise Exception("Word already exists in dictionary: "+word_text)
          for i in ALLREP_RANGE:
            self.from_ji_type[i].add(goi.kotoba(i),goi)
          self.nentries=self.nentries+1
        return self
        
    def addgoi(self,_kanji,_kana,_gaikoku,_initial_level,_kanji_atari=0,_kana_atari=0,_gaikoku_atari=0):
        self.add(Goi(self.initial_last_asked_event_number,_kanji,_kana,_gaikoku,_initial_level,_kanji_atari,_kana_atari,_gaikoku_atari))
        
    # Deletes the OPTIONAL_KIND representation of this word (i.e. removes its Kanji representation entirely)
    def remove_optional_representation(self,goi):
        self.from_ji_type[OPTIONAL_KIND].remove_representation(goi)
        
    def update_word(self,kind,goi,new_text):
        log("Updating",KIND_NAME[kind],'for',goi.pp(),'to',new_text) 
        # Plausibility check: new_text must not be in the goi pool (should be checked already by showarea) 
        if self.already_exists(kind,new_text):
            raise("Plausibility check violation. Word update generates duplicate word: "+new_text+"\n"+goi.pp())
        self.from_ji_type[kind].update(goi,new_text);
    
    def load_testdata(self):
        bark()
        self.addgoi("語学","ごがく","Linguistik, Sprachwissenschaft, Fach: Fremdsprache",3)
        self.addgoi("四度","よんど","Musik: Quart",3,1,-2,2)
        self.addgoi("","キセノン･ランプ","Xenonlampe",1,0,0,0)
        self.addgoi("旅行","りょこう","Reise",1,3,2,2)
        
    # Imports a dictionary from file pathname, either replacing the old one (merge is True) or
    # adding to the current one (merge is False).
    # Raises a String exception on error, which can be used to be displayed by the caller.
    # TODO: Wording of the exception should come from settings.
    # Format of Dictionary File:
    # First Line: KANREN01
    # Each entry consists of 4 lines:
    # A single colon
    # Representation in Kind 0 (or - (dash) if none exists) 
    # Representation in Kind 1
    # Representation in Kind 2
    # a comma-separated list of numbers with the following meaning:
    #    LEVEL for kind 0
    #    HIT COUNT for kind 0
    #   and so on for kind 1 and 2
    # or a single slash (/) indicating a new entry having no level and count yet (used for manually 
    # created dictionary files)
    # Leading and trailing spaces are ignored. Empty lines are ignored. The following symbols are
    # reserved and must not appear as the first or last character in a representation line:
    #    : ; + - / & = < >
    # The following symbols are reserved for later use and must not appear anywhere in the line:
    #    # \ @ $ 
    # Note that a # hash means that the rest of the line should be ignored, and that empty lines
    # are permitted too (and ignored). This is not used for creating the dictionary file, but only
    # when reading it.
    def dict_import(self,pathname,merge=False):
        # TODO: Proper exception handling. Maybe we should pass the exception to the caller, so
        #       that he can show an alert to the user. In this case, the exception text should
        #       come from the settings module.
        try:
            fileobj=CommentFileInput(pathname)
            if fileobj.eof():
                raise Exception("File "+pathname+" is empty")
        except IOError:
            e=sys.exc_info()[1]
            raise Exception("Can not import "+pathname+': '+e.strerror)
        else:
            log("Importing from",pathname," merge="+str(merge))
            version=fileobj.readline().strip()
            if version != DICT_VERSION:
                raise Exception("Incorrect dictionary version "+DICT_VERSION)
            errstr_filepart=" : file "+pathname+" line "
            fileobj.set_eof_exception("Premature end of file ("+pathname+")")
            representations=new_repr_list()
            new_dict_data=[]
            # default levels to be used for importing dictionaries using '/' to denote the level
            # information. If we are merging to an existing dictionary, we put those new words
            # to the current maximum level, considering them as difficult, until proved otherwise.
            default_number_information=new_repr_list()+new_repr_list() # for level and hit count
            defaultlevel=1
            for kind in ALLREP_RANGE:
                if merge and self.n_entries_for_kind(kind) > 0:
                    defaultlevel=self.from_ji_type[kind].maxlevel()        
                default_number_information[2*kind]=defaultlevel
                default_number_information[2*kind+1]=0
            while not fileobj.eof():
                line=fileobj.readline().rstrip() # single colon in column 1. Note that 
                                        # CommentFileInput.readline() already strips newlines, but
                                        # not trailing blanks.
                if line != ':':
                    raise "Not proper start of dictionary definition"+errstr_filepart+str(fileobj.lineno())
                for kind in ALLREP_RANGE:
                    repr=whitespace_normalize_word(fileobj.readlinex())
                    if(repr == REPRESENTATION_MISSING_STRING and kind == KANJI):
                        repr=''
                    errmsg=self.verify_word(repr,kind)
                    if len(errmsg) > 0:
                        raise errmsg+errstr_filepart+str(fileobj.lineno())
                    representations[kind]=repr
                num_info_line=fileobj.readlinex().strip()
                if num_info_line == '/':
                    number_information=default_number_information[:]
                else:
                  if not re.match(r'(\s*[+-]?\d+\s*,){5}\s*[+-]?\d+',num_info_line):
                    raise Exception("Level/HitCount specification invalid, must be comma-separated list of 6 numbers, but is '"+num_info_line+"'"+errstr_filepart+str(fileobj.lineno()))
                  number_information=list(map(lambda e: int(e),re.split(r'\s*,\s*',num_info_line)))
                  for kind in ALLREP_RANGE:
                    if number_information[2*kind] < 0:
                        raise "Negative level for "+KIND_NAME[kind]+errstr_filepart+str(fileobj.lineno())
                # Need shallow copy of representations, because this array is reused on every iteration
                new_dict_data.append([representations[:],number_information])
            if not merge:
                self.reset() # Erase existing dictionary
            for representations,number_information in new_dict_data:
                self.addgoi(representations[KANJI],representations[KANA],representations[GAIKOKUJI],number_information[0:2*REPRESENTATIONS:2],number_information[1],number_information[3], number_information[5])
            for i in ALLREP_RANGE:
                self.from_ji_type[i].level_compactification()
            if not merge:
                self.current_dict_pathname=pathname
        
    # Returns True if dictionary has been written, False otherwise
    def dict_export(self,pathname=''):
        if pathname=='':
            pathname=self.current_dict_pathname
        log("Exporting",self.n_entries_for_kind(GAIKOKUJI),"entries to",pathname)
        pathname_bak=pathname+'.bak'
        try:
            copyfile(pathname,pathname_bak)
        except IOError:
            # This exception is raised if the source file doesn't exist (yet), or if the destination
            # file is write protected
            e=sys.exc_info()[1]
            if(os.path.isfile(pathname)):
                showwarning("Access Problem","Can not create backup file for dictionary ("+pathname_bak+").\n"+e.strerror+"\nDictionary will be saved without being backed up.")                
        try:
            fileobj=uopen(pathname,'w')
        except IOError:
            e=sys.exc_info()[1]
            showerror("Access Problem","Can not write to dictionary file ("+pathname+").\n"+e.strerror)
            return False
        else:
            fileobj.write(DICT_VERSION+'\n')
            for goi in self.from_ji_type[GAIKOKUJI].each_goi():
                fileobj.write(':\n')
                number_information=[]
                log(goi.pp())
                for i in ALLREP_RANGE:
                    repr_for_kind=goi.gettext(i)
                    if repr_for_kind == '':
                        if i != OPTIONAL_KIND:
                            # Should we throw an exception here, or just report to the user?
                            todo("WARNING: Representation missing:",goi.pp())
                        fileobj.write(REPRESENTATION_MISSING_STRING+'\n')
                        for _ in range(0,2):
                            number_information.append('0')
                    else:
                        fileobj.write(repr_for_kind+'\n')
                        number_information.append(str(goi.kotoba(i).level()))
                        number_information.append(str(goi.kotoba(i).hit_count()))
                fileobj.write(','.join(number_information)+'\n')            
            fileobj.close
        if self.current_dict_pathname == '':
            self.current_dict_pathname=pathname
        return True
            
    def n_entries(self):
        return self.nentries
        
    def n_entries_for_kind(self,kind):
        return self.from_ji_type[kind].n_entries()
        
    def choose(self,kind,_event_history_number):
        return self.from_ji_type[kind].choose(_event_history_number,self.S.do_not_ask_again_steps())
        
    # Reorganise dictionary based on correctness of answer.
    # _goi : The Goi object in the challenge
    # _kind : The kind (i.e. KANJI), from where the challenge started
    # _answer_correct_p : true if the challenge was answered correctly, false otherwise
    # Returns: If a level change has resulted, returns the new level. Otherwise returns None.
    def answered_correctly(self,_goi,_kind,_answer_correct_p):
        level_change_info=_goi.adjust_hit_count(self.S,_kind,_answer_correct_p,self.from_ji_type[_kind].maximum_allowed_level(self.S))
        to_level=None
        if level_change_info != None:
            from_level=level_change_info
            if _answer_correct_p:
                to_level=from_level-1
            else:
                to_level=from_level+1
            self.from_ji_type[_kind].shift_level(_goi,from_level,to_level)
        return to_level
            
    def set_dirty(self):
        self.dirty=True
        
    def set_clean(self):
        self.dirty=False
        
    def is_dirty(self):
        return self.dirty
        
    # Verify a word coming from external source (user input), whether it doesn't contain invalid
    # characters. Returns error message if something invalid was found, or the empty string otherwise
    def verify_word(self,word,kind):
        if kind != OPTIONAL_KIND:
            if len(word) == 0:
                return self.S.msg__must_not_be_empty()
        if contains_forbidden_character(word):
            return self.S.msg__must_not_contain_these_characters()+REPRESENTAION_FORBIDDEN_CHARACTERS
        if has_contains_forbidden_character_at_border(word):
            return self.S.msg__border_must_not_contain_these_characters()+REPRESENTATION_FORBIDDEN_BORDER_CHARACTERS
        return ''
        
    def pp(self):
        # todo: Don't print, but return string (or let the caller choose on the output)
        prnt(self.current_dict_pp())
        prnt('is',bool_pp(self.is_dirty(),'dirty'))
        ne_str=' ('
        for kind in ALLREP_RANGE:
            ne_str=ne_str+' '+KIND_NAME[kind]+':'+str(self.n_entries_for_kind(kind))
        ne_str=ne_str+' )'
        prnt("Entries: ",self.n_entries(),ne_str)
        for kind in ALLREP_RANGE:
          prnt("From viewpoint",KIND_NAME[kind])
          for level in self.from_ji_type[kind].level_range():
            dict=self.from_ji_type[kind].translations(level)
            prnt("Level ",level,": Size = ",len(dict))
            for key in dict:
                goi=dict[key]
                prnt('  ',key,goi.kotoba(kind).hit_count(),goi)
        return self

if __name__ == "__main__":
    x=GoiDict()
    x.load_testdata()
    x.pp()
    prnt("Testing word choosing:")
    prnt(repr(x.choose(KANA)))
